#version 430

const int MAX_METABALLS = 17;
const float GRID_STEP = 0.1f;
const float ENERGY_THRESHOLD = 90.0f;
const float ROOT_MIN_THRESHOLD = 30.0f;
const float ROOT_MAX_THRESHOLD = 11000.0f;

layout (points) in;
layout (triangle_strip) out;
layout (max_vertices = 12) out;

layout(location = 1) uniform mat4 viewProjectionMatrix;
layout(location = 2) uniform vec3 metaballPositions[MAX_METABALLS];

out VertexData {
	vec3 position_world;
	vec3 normal_world;
	//vec2 uv;
} vert;

struct {
	bool inside;
	bool checked;
	int linkOne;
	int linkTwo;
	int linkThree;
	float energy;
} cubePointNodes[8] = {
{false, false, 1, 2, 4, 0.0f},
{false, false, 0, 5, 3, 0.0f},
{false, false, 0, 3, 6, 0.0f},
{false, false, 7, 2, 1, 0.0f},
{false, false, 0, 6, 5, 0.0f},
{false, false, 1, 4, 7, 0.0f},
{false, false, 2, 7, 4, 0.0f},
{false, false, 3, 5, 6, 0.0f}};

vec3 cubePoints[8];

int surfacePointCount;
vec3 surfacePoints[7];

//void calculateSurfacePoints(int parentIndex, int nodeIndex);

void main (void)
{
	// ###########################################
	// Render with a marching cube like algorithmn
	// ###########################################

	vec4 gridPoint = gl_in[0].gl_Position;
	cubePoints[0] = gridPoint.xyz;
	cubePoints[1] = vec3(gridPoint.x + GRID_STEP, gridPoint.y,				gridPoint.z);
	cubePoints[2] = vec3(gridPoint.x,				gridPoint.y,				gridPoint.z + GRID_STEP);
	cubePoints[3] = vec3(gridPoint.x + GRID_STEP, gridPoint.y,				gridPoint.z + GRID_STEP);
	cubePoints[4] = vec3(gridPoint.x,				gridPoint.y - GRID_STEP,	gridPoint.z);
	cubePoints[5] = vec3(gridPoint.x + GRID_STEP, gridPoint.y - GRID_STEP,	gridPoint.z);
	cubePoints[6] = vec3(gridPoint.x,				gridPoint.y - GRID_STEP,	gridPoint.z + GRID_STEP);
	cubePoints[7] = vec3(gridPoint.x + GRID_STEP, gridPoint.y - GRID_STEP,	gridPoint.z + GRID_STEP);
	
	
	int insideCounter = 0;
	float cumulativeEnergy;
	float distanceToMetaball;
	for(int i=0; i<8; i++) {
		cumulativeEnergy = 0.0f;
		for(int b=0; b < MAX_METABALLS; b++) {
			// +0.01f so that energy can not be infinite
			distanceToMetaball = length(cubePoints[i] - metaballPositions[b]) + 0.01f;
			cumulativeEnergy += 1/pow(distanceToMetaball,2);
		}
		
		// Check, if the whole grid will be all outside or all inside
		if(i == 0) {
			if(cumulativeEnergy < ROOT_MIN_THRESHOLD || cumulativeEnergy > ROOT_MAX_THRESHOLD) {
				return;
			}
		}
		
		if(cumulativeEnergy > ENERGY_THRESHOLD) {
			cubePointNodes[i].inside = true;
			insideCounter++;
		}
		cubePointNodes[i].energy = cumulativeEnergy;
	}
	if(insideCounter == 0 || insideCounter == 8) {
		return;
	}
	
	// calculate SurfacePoints
	// #######################
	/* for(int i=0; i<8; i++) {
		int surfacePointCount = 0;
		if(cubePointNodes[i].inside) {
			calculateSurfacePoints(i, i);
		}
		if(surfacePointCount > 2) {
			// emit surface points
		}
	} */
	
	
	vec3 normal;
	int frontIndex;
	int backIndex;
	
	int stackIndex = 0;
	int parentIndex;
	int nodeIndex;
	int parentIndexStack[8];
	int nodeIndexStack[8];
	int executionCount[8] = {0,0,0,0,0,0,0,0};
	for(int i=0; i<8; i++) {
		int surfacePointCount = 0;
		if(cubePointNodes[i].inside && !cubePointNodes[i].checked) {
			stackIndex = 0;
			parentIndexStack[stackIndex] = i;
			nodeIndexStack[stackIndex] = i;
			while(stackIndex >= 0) {
				parentIndex = parentIndexStack[stackIndex];
				nodeIndex = nodeIndexStack[stackIndex];
				if(cubePointNodes[nodeIndex].inside) {
					cubePointNodes[nodeIndex].checked = true;
					executionCount[nodeIndex]++;
					if(parentIndex == nodeIndex) {
						switch(executionCount[nodeIndex]) {
							case 1:
								//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkOne);
								if(executionCount[cubePointNodes[nodeIndex].linkOne] == 0) {
									stackIndex++;
									parentIndexStack[stackIndex] = nodeIndex;
									nodeIndexStack[stackIndex] = cubePointNodes[nodeIndex].linkOne;
									continue;
								} else {
									// Means, that the node was already called and checked=true
									// Pass through to next execution
									executionCount[nodeIndex]++;
								}
							case 2:
								//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkTwo);
								if(executionCount[cubePointNodes[nodeIndex].linkTwo] == 0) {
									stackIndex++;
									parentIndexStack[stackIndex] = nodeIndex;
									nodeIndexStack[stackIndex] = cubePointNodes[nodeIndex].linkTwo;
									continue;
								} else {
									// Means, that the node was already called and checked=true
									// Pass through to next execution
									executionCount[nodeIndex]++;
								}
							case 3:
								//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkThree);
								if(executionCount[cubePointNodes[nodeIndex].linkThree] == 0) {
									stackIndex++;
									parentIndexStack[stackIndex] = nodeIndex;
									nodeIndexStack[stackIndex] = cubePointNodes[nodeIndex].linkThree;
									continue;
								} else {
									// Means, that the node was already called and checked=true
									// Pass through to next execution
									executionCount[nodeIndex]++;
								}
						}
					} else if(cubePointNodes[nodeIndex].linkOne == parentIndex) {
						switch(executionCount[nodeIndex]) {
							case 1:
								//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkTwo);
								if(executionCount[cubePointNodes[nodeIndex].linkTwo] == 0) {
									stackIndex++;
									parentIndexStack[stackIndex] = nodeIndex;
									nodeIndexStack[stackIndex] = cubePointNodes[nodeIndex].linkTwo;
									continue;
								} else {
									// Means, that the node was already called and checked=true
									// Pass through to next execution
									executionCount[nodeIndex]++;
								}
							case 2:
								//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkThree);
								if(executionCount[cubePointNodes[nodeIndex].linkThree] == 0) {
									stackIndex++;
									parentIndexStack[stackIndex] = nodeIndex;
									nodeIndexStack[stackIndex] = cubePointNodes[nodeIndex].linkThree;
									continue;
								} else {
									// Means, that the node was already called and checked=true
									// Pass through to next execution
									executionCount[nodeIndex]++;
								}
						}
					} else if(cubePointNodes[nodeIndex].linkTwo == parentIndex) {
						switch(executionCount[nodeIndex]) {
							case 1:
								//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkThree);
								if(executionCount[cubePointNodes[nodeIndex].linkThree] == 0) {
									stackIndex++;
									parentIndexStack[stackIndex] = nodeIndex;
									nodeIndexStack[stackIndex] = cubePointNodes[nodeIndex].linkThree;
									continue;
								} else {
									// Means, that the node was already called and checked=true
									// Pass through to next execution
									executionCount[nodeIndex]++;
								}
							case 2:
								//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkOne);
								if(executionCount[cubePointNodes[nodeIndex].linkOne] == 0) {
									stackIndex++;
									parentIndexStack[stackIndex] = nodeIndex;
									nodeIndexStack[stackIndex] = cubePointNodes[nodeIndex].linkOne;
									continue;
								} else {
									// Means, that the node was already called and checked=true
									// Pass through to next execution
									executionCount[nodeIndex]++;
								}
						}
					} else if(cubePointNodes[nodeIndex].linkThree == parentIndex) {
						switch(executionCount[nodeIndex]) {
							case 1:
								//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkOne);
								if(executionCount[cubePointNodes[nodeIndex].linkOne] == 0) {
									stackIndex++;
									parentIndexStack[stackIndex] = nodeIndex;
									nodeIndexStack[stackIndex] = cubePointNodes[nodeIndex].linkOne;
									continue;
								} else {
									// Means, that the node was already called and checked=true
									// Pass through to next execution
									executionCount[nodeIndex]++;
								}
							case 2:
								//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkTwo);
								if(executionCount[cubePointNodes[nodeIndex].linkTwo] == 0) {
									stackIndex++;
									parentIndexStack[stackIndex] = nodeIndex;
									nodeIndexStack[stackIndex] = cubePointNodes[nodeIndex].linkTwo;
									continue;
								} else {
									// Means, that the node was already called and checked=true
									// Pass through to next execution
									executionCount[nodeIndex]++;
								}
						}
					}
				} else {
					float ratio = (ENERGY_THRESHOLD - cubePointNodes[nodeIndex].energy) / (cubePointNodes[parentIndex].energy - cubePointNodes[nodeIndex].energy);
					surfacePoints[surfacePointCount++] = cubePoints[parentIndex]*ratio + cubePoints[nodeIndex]*(1-ratio);
				}
				stackIndex--;
			}
			
			if(surfacePointCount > 2) {
				// emit surface points
				
				// calculateNormal
				normal = normalize(cross(surfacePoints[1] - surfacePoints[0], surfacePoints[2] - surfacePoints[0]));
				
				// for a triangle_strip take one from the front and one from the back of the pointArray toggling
				frontIndex = 0;
				backIndex = surfacePointCount-1;
				
				// start with index 0 - then toggle front/back
				vert.position_world = surfacePoints[frontIndex++];
				vert.normal_world = normal;
				gl_Position = viewProjectionMatrix * vec4(vert.position_world, 1.0f);
				EmitVertex();
				
				while(frontIndex <= backIndex) {
					vert.position_world = surfacePoints[frontIndex++];
					vert.normal_world = normal;
					gl_Position = viewProjectionMatrix * vec4(vert.position_world, 1.0f);
					EmitVertex();
					if(frontIndex <= backIndex) {
						vert.position_world = surfacePoints[backIndex--];
						vert.normal_world = normal;
						gl_Position = viewProjectionMatrix * vec4(vert.position_world, 1.0f);
						EmitVertex();
					}
				}
				
				EndPrimitive();
			}
		}
	}
}

/*
void calculateSurfacePoints(int parentIndex, int nodeIndex) {
	if(cubePointNodes[nodeIndex].checked) {
		return;
	}
	if(cubePointNodes[nodeIndex].inside) {
		cubePointNodes[nodeIndex].checked = true;
		if(parentIndex == nodeIndex) {
			//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkOne);
			//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkTwo);
			//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkThree);
		} else if(cubePointNodes[nodeIndex].linkOne == parentIndex) {
			//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkTwo);
			//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkThree);
		} else if(cubePointNodes[nodeIndex].linkTwo == parentIndex) {
			//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkThree);
			//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkOne);
		} else if(cubePointNodes[nodeIndex].linkThree == parentIndex) {
			//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkOne);
			//calculateSurfacePoints(nodeIndex, cubePointNodes[nodeIndex].linkTwo);
		}
	} else {
		float ratio = (ENERGY_THRESHOLD - cubePointNodes[nodeIndex].energy) / (cubePointNodes[parentIndex].energy - cubePointNodes[nodeIndex].energy);
		surfacePoints[surfacePointCount++] = cubePoints[parentIndex]*ratio + cubePoints[nodeIndex]*(1-ratio);
	}
}
*/