#version 430 core
/*
* Copyright 2010 Vienna University of Technology.
* Institute of Computer Graphics and Algorithms.
* This file is part of the ECG Lab Framework and must not be redistributed.
*/

const int MAX_BONES = 20;
const int MAX_WEIGHTS = 4;

layout(location = 0) in vec3 position;
layout(location = 1) in vec3 normal;
layout(location = 2) in vec2 uv;

layout(location = 3) in ivec4 boneWeightIndices;
layout(location = 4) in vec4 boneWeights;

out VertexData {
	vec3 position_world;
	vec3 normal_world;
	vec2 uv;
} vert;

uniform mat4 modelMatrix;
uniform mat4 viewProjMatrix;
uniform mat3 normalMatrix;
layout(location = 0) uniform mat4 boneMatrices[MAX_BONES];

void main() {
	//vec4 positionSum = vec4(position, 1.0f);
	//vec4 normalSum = vec4(normal, 1.0f);
	
	vec4 positionSum = vec4(0.0f, 0.0f, 0.0f, 0.0f);
	vec4 normalSum = vec4(0.0f, 0.0f, 0.0f, 0.0f);
	for(int i=0; i<MAX_WEIGHTS; i++) {
		if(boneWeightIndices[i] == -1) {
			continue;
		}
		positionSum += boneMatrices[boneWeightIndices[i]] * vec4(position, 1) * boneWeights[i];
		normalSum += boneMatrices[boneWeightIndices[i]] * vec4(normal, 1) * boneWeights[i];
	}
	positionSum.w = 1.0f;
	normalSum.w = 1.0f;

	vert.normal_world = normalMatrix * normalSum.xyz;
	vert.uv = uv;
	vec4 position_world_ = modelMatrix * positionSum;
	vert.position_world = position_world_.xyz;
	gl_Position = viewProjMatrix * position_world_;
}