#version 430 core
/*
* Copyright 2010 Vienna University of Technology.
* Institute of Computer Graphics and Algorithms.
* This file is part of the ECG Lab Framework and must not be redistributed.
* parts of this were taken from https://www.youtube.com/watch?v=5yhDb9dzJ58
*/

layout(location = 0) in vec3 position;
layout(location = 1) in vec3 normal;
layout(location = 2) in vec2 uv;

out VertexData {
	vec3 position_world;
	vec3 normal_world;
	vec2 uv;
    vec3 normal_verschiebung;
} vert;

uniform mat4 modelMatrix;
uniform mat4 viewProjMatrix;
uniform mat3 normalMatrix;
uniform float time;

const float PI = 3.1415926535897932384626433832795;

const float waveLength = 10.0f;
const float waveAmplitude = 0.5f;

float generateOffset(float x, float z){
	float radiansX = ((mod(x+z*x, waveLength)/waveLength) + time * mod(x * 0.8 + z, 1.5)) * 2.0 * PI;
	float radiansZ = ((mod(z*x +x*z, waveLength)/waveLength) + time * 2.0 * mod(x , 2.0) ) * 2.0 * PI;
	return waveAmplitude * 0.5f * (sin(radiansZ) + cos(radiansX));
}

void main() {
    float verschiebung = generateOffset(uv[0], uv[1]);
	vert.normal_world = normalMatrix * normal;

	vert.uv = uv;

	//vec4 newPos = modelMatrix * vec4(position + vert.normal_verschiebung, 1);
	vec4 newPos = modelMatrix * vec4(position + verschiebung, 1);
	vert.position_world = newPos.xyz;

	gl_Position = viewProjMatrix * newPos;
}